import decimal
import typing as t
from unittest import TestCase

import pydantic
import pydantic_core
import pytest
from pydantic import ValidationError

from piccolo.columns import (
    JSON,
    JSONB,
    UUID,
    Array,
    Email,
    Integer,
    Numeric,
    Secret,
    Text,
    Time,
    Timestamp,
    Timestamptz,
    Varchar,
)
from piccolo.columns.column_types import ForeignKey
from piccolo.table import Table
from piccolo.utils.pydantic import create_pydantic_model


class TestVarcharColumn(TestCase):
    def test_varchar_length(self):
        class Director(Table):
            name = Varchar(length=10)

        pydantic_model = create_pydantic_model(table=Director)

        with self.assertRaises(ValidationError):
            pydantic_model(name="This is a really long name")

        pydantic_model(name="short name")


class TestEmailColumn(TestCase):
    def test_email(self):
        class Director(Table):
            email = Email()

        pydantic_model = create_pydantic_model(table=Director)

        self.assertEqual(
            pydantic_model.model_json_schema()["properties"]["email"]["anyOf"][
                0
            ]["format"],
            "email",
        )

        with self.assertRaises(ValidationError):
            pydantic_model(email="not a valid email")

        # Shouldn't raise an exception:
        pydantic_model(email="test@gmail.com")


class TestNumericColumn(TestCase):
    """
    Numeric and Decimal are the same - so we'll just test Numeric.
    """

    def test_numeric_digits(self):
        class Movie(Table):
            box_office = Numeric(digits=(5, 1))

        pydantic_model = create_pydantic_model(table=Movie)

        with self.assertRaises(ValidationError):
            # This should fail as there are too much numbers after the decimal
            # point
            pydantic_model(box_office=decimal.Decimal("1.11"))

        with self.assertRaises(ValidationError):
            # This should fail as there are too much numbers in total
            pydantic_model(box_office=decimal.Decimal("11111.1"))

        pydantic_model(box_office=decimal.Decimal("1.0"))

    def test_numeric_without_digits(self):
        class Movie(Table):
            box_office = Numeric()

        try:
            create_pydantic_model(table=Movie)
        except TypeError:
            self.fail(
                "Creating numeric field without"
                " digits failed in pydantic model."
            )
        else:
            self.assertTrue(True)


class TestSecretColumn(TestCase):
    def test_secret_param(self):
        class TopSecret(Table):
            confidential = Secret()

        pydantic_model = create_pydantic_model(table=TopSecret)
        self.assertEqual(
            pydantic_model.model_json_schema()["properties"]["confidential"][
                "extra"
            ]["secret"],
            True,
        )


class TestArrayColumn(TestCase):
    def test_array_param(self):
        class Band(Table):
            members = Array(base_column=Varchar(length=16))

        pydantic_model = create_pydantic_model(table=Band)

        self.assertEqual(
            pydantic_model.model_json_schema()["properties"]["members"][
                "anyOf"
            ][0]["items"]["type"],
            "string",
        )

    def test_multidimensional_array(self):
        """
        Make sure that multidimensional arrays have the correct type.
        """

        class Band(Table):
            members = Array(Array(Varchar(length=255)), required=True)

        pydantic_model = create_pydantic_model(table=Band)

        self.assertEqual(
            pydantic_model.model_fields["members"].annotation,
            t.List[t.List[pydantic.constr(max_length=255)]],
        )

        # Should not raise a validation error:
        pydantic_model(
            members=[
                ["Alice", "Bob", "Francis"],
                ["Alan", "Georgia", "Sue"],
            ]
        )

        with self.assertRaises(ValueError):
            pydantic_model(members=["Bob"])


class TestForeignKeyColumn(TestCase):
    def test_target_column(self):
        """
        Make sure the `target_column` is correctly set in the Pydantic schema
        for `ForeignKey` columns.
        """

        class Manager(Table):
            name = Varchar(unique=True)

        class BandA(Table):
            manager = ForeignKey(Manager, target_column=Manager.name)

        class BandB(Table):
            manager = ForeignKey(Manager, target_column="name")

        class BandC(Table):
            manager = ForeignKey(Manager)

        self.assertEqual(
            create_pydantic_model(table=BandA).model_json_schema()[
                "properties"
            ]["manager"]["extra"]["foreign_key"]["target_column"],
            "name",
        )

        self.assertEqual(
            create_pydantic_model(table=BandB).model_json_schema()[
                "properties"
            ]["manager"]["extra"]["foreign_key"]["target_column"],
            "name",
        )

        self.assertEqual(
            create_pydantic_model(table=BandC).model_json_schema()[
                "properties"
            ]["manager"]["extra"]["foreign_key"]["target_column"],
            "id",
        )


class TestTextColumn(TestCase):
    def test_text_widget(self):
        """
        Make sure that we indicate that `Text` columns require a special widget
        in Piccolo Admin.
        """

        class Band(Table):
            bio = Text()

        pydantic_model = create_pydantic_model(table=Band)

        self.assertEqual(
            pydantic_model.model_json_schema()["properties"]["bio"]["extra"][
                "widget"
            ],
            "text-area",
        )


class TestTimeColumn(TestCase):
    def test_time_format(self):
        class Concert(Table):
            start_time = Time()

        pydantic_model = create_pydantic_model(table=Concert)

        self.assertEqual(
            pydantic_model.model_json_schema()["properties"]["start_time"][
                "anyOf"
            ][0]["format"],
            "time",
        )


class TestTimestamptzColumn(TestCase):
    def test_timestamptz_widget(self):
        """
        Make sure that we indicate that `Timestamptz` columns require a special
        widget in Piccolo Admin.
        """

        class Concert(Table):
            starts_on_1 = Timestamptz()
            starts_on_2 = Timestamp()

        pydantic_model = create_pydantic_model(table=Concert)

        properties = pydantic_model.model_json_schema()["properties"]

        self.assertEqual(
            properties["starts_on_1"]["extra"]["widget"],
            "timestamptz",
        )

        self.assertIsNone(properties["starts_on_2"]["extra"].get("widget"))


class TestUUIDColumn(TestCase):
    class Ticket(Table):
        code = UUID()

    def setUp(self):
        self.Ticket.create_table().run_sync()

    def tearDown(self):
        self.Ticket.alter().drop_table().run_sync()

    def test_uuid_format(self):
        class Ticket(Table):
            code = UUID()

        pydantic_model = create_pydantic_model(table=Ticket)

        ticket = Ticket()
        ticket.save().run_sync()

        # We'll also fetch it from the DB in case the database adapter's UUID
        # is used.
        ticket_from_db = Ticket.objects().first().run_sync()
        assert ticket_from_db is not None

        for ticket_ in (ticket, ticket_from_db):
            json = pydantic_model(**ticket_.to_dict()).model_dump_json()
            self.assertEqual(json, '{"code":"' + str(ticket_.code) + '"}')

        self.assertEqual(
            pydantic_model.model_json_schema()["properties"]["code"]["anyOf"][
                0
            ]["format"],
            "uuid",
        )


class TestColumnHelpText(TestCase):
    """
    Make sure that columns with `help_text` attribute defined have the
    relevant text appear in the schema.
    """

    def test_column_help_text_present(self):
        help_text = "In millions of US dollars."

        class Movie(Table):
            box_office = Numeric(digits=(5, 1), help_text=help_text)

        pydantic_model = create_pydantic_model(table=Movie)

        self.assertEqual(
            pydantic_model.model_json_schema()["properties"]["box_office"][
                "extra"
            ]["help_text"],
            help_text,
        )


class TestTableHelpText(TestCase):
    """
    Make sure that tables with `help_text` attribute defined have the
    relevant text appear in the schema.
    """

    def test_table_help_text_present(self):
        help_text = "Movies which were released in cinemas."

        class Movie(Table, help_text=help_text):
            name = Varchar()

        pydantic_model = create_pydantic_model(table=Movie)

        self.assertEqual(
            pydantic_model.model_json_schema()["extra"]["help_text"],
            help_text,
        )


class TestUniqueColumn(TestCase):
    def test_unique_column_true(self):
        class Director(Table):
            name = Varchar(unique=True)

        pydantic_model = create_pydantic_model(table=Director)

        self.assertEqual(
            pydantic_model.model_json_schema()["properties"]["name"]["extra"][
                "unique"
            ],
            True,
        )

    def test_unique_column_false(self):
        class Director(Table):
            name = Varchar()

        pydantic_model = create_pydantic_model(table=Director)

        self.assertEqual(
            pydantic_model.model_json_schema()["properties"]["name"]["extra"][
                "unique"
            ],
            False,
        )


class TestJSONColumn(TestCase):
    def test_default(self):
        class Movie(Table):
            meta = JSON()
            meta_b = JSONB()

        pydantic_model = create_pydantic_model(table=Movie)

        json_string = '{"code": 12345}'

        model_instance = pydantic_model(meta=json_string, meta_b=json_string)
        self.assertEqual(model_instance.meta, json_string)  # type: ignore
        self.assertEqual(model_instance.meta_b, json_string)  # type: ignore

    def test_deserialize_json(self):
        class Movie(Table):
            meta = JSON()
            meta_b = JSONB()

        pydantic_model = create_pydantic_model(
            table=Movie, deserialize_json=True
        )

        json_string = '{"code": 12345}'
        output = {"code": 12345}

        model_instance = pydantic_model(meta=json_string, meta_b=json_string)
        self.assertEqual(model_instance.meta, output)  # type: ignore
        self.assertEqual(model_instance.meta_b, output)  # type: ignore

    def test_validation(self):
        class Movie(Table):
            meta = JSON()
            meta_b = JSONB()

        for deserialize_json in (True, False):
            pydantic_model = create_pydantic_model(
                table=Movie, deserialize_json=deserialize_json
            )

            json_string = "error"

            with self.assertRaises(pydantic.ValidationError):
                pydantic_model(meta=json_string, meta_b=json_string)

    def test_json_widget(self):
        """
        Make sure that we indicate that `JSON` / `JSONB` columns require a
        special widget in Piccolo Admin.
        """

        class Movie(Table):
            features = JSON()

        pydantic_model = create_pydantic_model(table=Movie)

        self.assertEqual(
            pydantic_model.model_json_schema()["properties"]["features"][
                "extra"
            ]["widget"],
            "json",
        )

    def test_null_value(self):
        class Movie(Table):
            meta = JSON(null=True)
            meta_b = JSONB(null=True)

        pydantic_model = create_pydantic_model(table=Movie)
        movie = pydantic_model(meta=None, meta_b=None)

        self.assertIsNone(movie.meta)  # type: ignore
        self.assertIsNone(movie.meta_b)  # type: ignore


class TestExcludeColumns(TestCase):
    def test_all(self):
        class Computer(Table):
            CPU = Varchar()
            GPU = Varchar()

        pydantic_model = create_pydantic_model(Computer, exclude_columns=())

        properties = pydantic_model.model_json_schema()["properties"]
        self.assertIsInstance(properties["GPU"], dict)
        self.assertIsInstance(properties["CPU"], dict)

    def test_exclude(self):
        class Computer(Table):
            CPU = Varchar()
            GPU = Varchar()

        pydantic_model = create_pydantic_model(
            Computer,
            exclude_columns=(Computer.CPU,),
        )

        properties = pydantic_model.model_json_schema()["properties"]
        self.assertIsInstance(properties.get("GPU"), dict)
        self.assertIsNone(properties.get("CPU"))

    def test_exclude_all_manually(self):
        class Computer(Table):
            GPU = Varchar()
            CPU = Varchar()

        pydantic_model = create_pydantic_model(
            Computer,
            exclude_columns=(Computer.GPU, Computer.CPU),
        )

        self.assertEqual(pydantic_model.model_json_schema()["properties"], {})

    def test_exclude_all_meta(self):
        class Computer(Table):
            GPU = Varchar()
            CPU = Varchar()

        pydantic_model = create_pydantic_model(
            Computer,
            exclude_columns=tuple(Computer._meta.columns),
        )

        self.assertEqual(pydantic_model.model_json_schema()["properties"], {})

    def test_invalid_column_str(self):
        class Computer(Table):
            CPU = Varchar()
            GPU = Varchar()

        with self.assertRaises(ValueError):
            create_pydantic_model(
                Computer,
                exclude_columns=("CPU",),  # type: ignore
            )

    def test_invalid_column_different_table(self):
        class Computer(Table):
            CPU = Varchar()
            GPU = Varchar()

        class Computer2(Table):
            SSD = Varchar()

        with self.assertRaises(ValueError):
            create_pydantic_model(Computer, exclude_columns=(Computer2.SSD,))

    def test_invalid_column_different_table_same_type(self):
        class Computer(Table):
            CPU = Varchar()
            GPU = Varchar()

        class Computer2(Table):
            CPU = Varchar()

        with self.assertRaises(ValueError):
            create_pydantic_model(Computer, exclude_columns=(Computer2.CPU,))

    def test_exclude_nested(self):
        class Manager(Table):
            name = Varchar()
            phone_number = Integer()

        class Band(Table):
            name = Varchar()
            manager = ForeignKey(Manager)
            popularity = Integer()

        pydantic_model = create_pydantic_model(
            table=Band,
            exclude_columns=(
                Band.popularity,
                Band.manager.phone_number,
            ),
            nested=(Band.manager,),
        )

        model_instance = pydantic_model(
            name="Pythonistas", manager={"name": "Guido"}
        )
        self.assertEqual(
            model_instance.model_dump(),
            {"name": "Pythonistas", "manager": {"name": "Guido"}},
        )


class TestIncludeColumns(TestCase):
    def test_include(self):
        class Band(Table):
            name = Varchar()
            popularity = Integer()

        pydantic_model = create_pydantic_model(
            Band,
            include_columns=(Band.name,),
        )

        properties = pydantic_model.model_json_schema()["properties"]
        self.assertIsInstance(properties.get("name"), dict)
        self.assertIsNone(properties.get("popularity"))

    def test_include_exclude_error(self):
        """
        An exception should be raised if both `include_columns` and
        `exclude_columns` are provided.
        """

        class Band(Table):
            name = Varchar()
            popularity = Integer()

        with self.assertRaises(ValueError):
            create_pydantic_model(
                Band,
                exclude_columns=(Band.name,),
                include_columns=(Band.name,),
            )

    def test_nested(self):
        """
        Make sure that columns on related tables work.
        """

        class Manager(Table):
            name = Varchar()
            phone_number = Integer()

        class Band(Table):
            name = Varchar()
            manager = ForeignKey(Manager)
            popularity = Integer()

        pydantic_model = create_pydantic_model(
            table=Band,
            include_columns=(
                Band.name,
                Band.manager.name,
            ),
            nested=(Band.manager,),
        )

        model_instance = pydantic_model(
            name="Pythonistas", manager={"name": "Guido"}
        )
        self.assertEqual(
            model_instance.model_dump(),
            {"name": "Pythonistas", "manager": {"name": "Guido"}},
        )


class TestNestedModel(TestCase):
    def test_true(self):
        """
        Make sure all foreign key columns are converted to nested models, when
        `nested=True`.
        """

        class Country(Table):
            name = Varchar(length=10)

        class Manager(Table):
            name = Varchar(length=10)
            country = ForeignKey(Country)

        class Band(Table):
            name = Varchar(length=10)
            manager = ForeignKey(Manager)

        BandModel = create_pydantic_model(table=Band, nested=True)

        #######################################################################

        ManagerModel = t.cast(
            t.Type[pydantic.BaseModel],
            BandModel.model_fields["manager"].annotation,
        )
        self.assertTrue(issubclass(ManagerModel, pydantic.BaseModel))
        self.assertEqual(
            [i for i in ManagerModel.model_fields.keys()], ["name", "country"]
        )

        #######################################################################

        CountryModel = t.cast(
            t.Type[pydantic.BaseModel],
            ManagerModel.model_fields["country"].annotation,
        )
        self.assertTrue(issubclass(CountryModel, pydantic.BaseModel))
        self.assertEqual(
            [i for i in CountryModel.model_fields.keys()], ["name"]
        )

    def test_tuple(self):
        """
        Make sure only the specified foreign key columns are converted to
        nested models.
        """

        class Country(Table):
            name = Varchar()

        class Manager(Table):
            name = Varchar()
            country = ForeignKey(Country)

        class Band(Table):
            name = Varchar()
            manager = ForeignKey(Manager)
            assistant_manager = ForeignKey(Manager)

        class Venue(Table):
            name = Varchar()

        class Concert(Table):
            band_1 = ForeignKey(Band)
            band_2 = ForeignKey(Band)
            venue = ForeignKey(Venue)

        #######################################################################
        # Test one level deep

        BandModel = create_pydantic_model(table=Band, nested=(Band.manager,))

        ManagerModel = t.cast(
            t.Type[pydantic.BaseModel],
            BandModel.model_fields["manager"].annotation,
        )
        self.assertTrue(issubclass(ManagerModel, pydantic.BaseModel))
        self.assertEqual(
            [i for i in ManagerModel.model_fields.keys()], ["name", "country"]
        )
        self.assertEqual(ManagerModel.__qualname__, "Band.manager")

        AssistantManagerType = BandModel.model_fields[
            "assistant_manager"
        ].annotation
        self.assertIs(AssistantManagerType, t.Optional[int])

        #######################################################################
        # Test two levels deep

        BandModel = create_pydantic_model(
            table=Band, nested=(Band.manager._.country,)
        )

        ManagerModel = t.cast(
            t.Type[pydantic.BaseModel],
            BandModel.model_fields["manager"].annotation,
        )
        self.assertTrue(issubclass(ManagerModel, pydantic.BaseModel))
        self.assertEqual(
            [i for i in ManagerModel.model_fields.keys()], ["name", "country"]
        )
        self.assertEqual(ManagerModel.__qualname__, "Band.manager")

        AssistantManagerType = t.cast(
            t.Type[pydantic.BaseModel],
            BandModel.model_fields["assistant_manager"].annotation,
        )
        self.assertIs(AssistantManagerType, t.Optional[int])

        CountryModel = t.cast(
            t.Type[pydantic.BaseModel],
            ManagerModel.model_fields["country"].annotation,
        )
        self.assertTrue(issubclass(CountryModel, pydantic.BaseModel))
        self.assertEqual(
            [i for i in CountryModel.model_fields.keys()], ["name"]
        )
        self.assertEqual(CountryModel.__qualname__, "Band.manager.country")

        #######################################################################
        # Test three levels deep

        ConcertModel = create_pydantic_model(
            Concert, nested=(Concert.band_1._.manager,)
        )

        VenueModel = ConcertModel.model_fields["venue"].annotation
        self.assertIs(VenueModel, t.Optional[int])

        BandModel = t.cast(
            t.Type[pydantic.BaseModel],
            ConcertModel.model_fields["band_1"].annotation,
        )
        self.assertTrue(issubclass(BandModel, pydantic.BaseModel))
        self.assertEqual(
            [i for i in BandModel.model_fields.keys()],
            ["name", "manager", "assistant_manager"],
        )
        self.assertEqual(BandModel.__qualname__, "Concert.band_1")

        ManagerModel = t.cast(
            t.Type[pydantic.BaseModel],
            BandModel.model_fields["manager"].annotation,
        )
        self.assertTrue(issubclass(ManagerModel, pydantic.BaseModel))
        self.assertEqual(
            [i for i in ManagerModel.model_fields.keys()],
            ["name", "country"],
        )
        self.assertEqual(ManagerModel.__qualname__, "Concert.band_1.manager")

        AssistantManagerType = BandModel.model_fields[
            "assistant_manager"
        ].annotation
        self.assertIs(AssistantManagerType, t.Optional[int])

        CountryModel = ManagerModel.model_fields["country"].annotation
        self.assertIs(CountryModel, t.Optional[int])

        #######################################################################
        # Test with `model_name` arg

        MyConcertModel = create_pydantic_model(
            Concert,
            nested=(Concert.band_1._.manager,),
            model_name="MyConcertModel",
        )

        BandModel = t.cast(
            t.Type[pydantic.BaseModel],
            MyConcertModel.model_fields["band_1"].annotation,
        )
        self.assertEqual(BandModel.__qualname__, "MyConcertModel.band_1")

        ManagerModel = BandModel.model_fields["manager"].annotation
        self.assertEqual(
            ManagerModel.__qualname__, "MyConcertModel.band_1.manager"
        )

    def test_cascaded_args(self) -> None:
        """
        Make sure that arguments passed to ``create_pydantic_model`` are
        cascaded to nested models.
        """

        class Country(Table):
            name = Varchar(length=10)

        class Manager(Table):
            name = Varchar(length=10)
            country = ForeignKey(Country)

        class Band(Table):
            name = Varchar(length=10)
            manager = ForeignKey(Manager)

        BandModel = create_pydantic_model(
            table=Band, nested=True, include_default_columns=True
        )

        ManagerModel = t.cast(
            t.Type[pydantic.BaseModel],
            BandModel.model_fields["manager"].annotation,
        )
        self.assertTrue(issubclass(ManagerModel, pydantic.BaseModel))
        self.assertEqual(
            [i for i in ManagerModel.model_fields.keys()],
            ["id", "name", "country"],
        )

        CountryModel = t.cast(
            t.Type[pydantic.BaseModel],
            ManagerModel.model_fields["country"].annotation,
        )
        self.assertTrue(issubclass(CountryModel, pydantic.BaseModel))
        self.assertEqual(
            [i for i in CountryModel.model_fields.keys()], ["id", "name"]
        )


class TestRecursionDepth(TestCase):
    def test_max(self):
        class Country(Table):
            name = Varchar()

        class Manager(Table):
            name = Varchar()
            country = ForeignKey(Country)

        class Band(Table):
            name = Varchar()
            manager = ForeignKey(Manager)
            assistant_manager = ForeignKey(Manager)

        class Venue(Table):
            name = Varchar()

        class Concert(Table):
            band = ForeignKey(Band)
            venue = ForeignKey(Venue)

        ConcertModel = create_pydantic_model(
            table=Concert, nested=True, max_recursion_depth=2
        )

        VenueModel = t.cast(
            t.Type[pydantic.BaseModel],
            ConcertModel.model_fields["venue"].annotation,
        )
        self.assertTrue(issubclass(VenueModel, pydantic.BaseModel))

        BandModel = t.cast(
            t.Type[pydantic.BaseModel],
            ConcertModel.model_fields["band"].annotation,
        )
        self.assertTrue(issubclass(BandModel, pydantic.BaseModel))

        ManagerModel = t.cast(
            t.Type[pydantic.BaseModel],
            BandModel.model_fields["manager"].annotation,
        )
        self.assertTrue(issubclass(ManagerModel, pydantic.BaseModel))

        # We should have hit the recursion depth:
        CountryModel = ManagerModel.model_fields["country"].annotation
        self.assertIs(CountryModel, t.Optional[int])


class TestDBColumnName(TestCase):
    def test_db_column_name(self):
        """
        Make sure that the Pydantic model has an alias if ``db_column_name``
        is specified for a column.
        """

        class Band(Table):
            name = Varchar(db_column_name="regrettable_column_name")

        BandModel = create_pydantic_model(table=Band)

        model = BandModel(regrettable_column_name="test")

        self.assertEqual(model.name, "test")  # type: ignore


class TestJSONSchemaExtra(TestCase):
    def test_json_schema_extra(self):
        """
        Make sure that the ``json_schema_extra`` arguments are reflected in
        Pydantic model's schema.
        """

        class Band(Table):
            name = Varchar()

        model = create_pydantic_model(
            Band, json_schema_extra={"extra": {"visible_columns": ("name",)}}
        )
        self.assertEqual(
            model.model_json_schema()["extra"]["visible_columns"], ("name",)
        )


class TestPydanticExtraFields(TestCase):
    def test_pydantic_extra_fields(self) -> None:
        """
        Make sure that the value of ``extra`` in the config class
        is correctly propagated to the generated model.
        """

        class Band(Table):
            name = Varchar()

        config: pydantic.config.ConfigDict = {"extra": "forbid"}
        model = create_pydantic_model(Band, pydantic_config=config)

        self.assertEqual(model.model_config.get("extra"), "forbid")

    def test_pydantic_invalid_extra_fields(self) -> None:
        """
        Make sure that invalid values for ``extra`` in the config class
        are rejected.
        """

        class Band(Table):
            name = Varchar()

        config: pydantic.config.ConfigDict = {
            "extra": "foobar"  # type: ignore
        }

        with pytest.raises(pydantic_core._pydantic_core.SchemaError):
            create_pydantic_model(Band, pydantic_config=config)



from piccolo.apps.migrations.auto import MigrationManager
from piccolo.columns.column_types import Boolean, Timestamp
from piccolo.columns.indexes import IndexMethod

ID = "2021-04-30T16:14:15"
VERSION = "0.18.2"


async def forwards():
    manager = MigrationManager(migration_id=ID, app_name="user")

    manager.add_column(
        table_class_name="BaseUser",
        tablename="piccolo_user",
        column_name="superuser",
        column_class_name="Boolean",
        column_class=Boolean,
        params={
            "default": False,
            "null": False,
            "primary": False,
            "key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
        },
    )

    manager.add_column(
        table_class_name="BaseUser",
        tablename="piccolo_user",
        column_name="last_login",
        column_class_name="Timestamp",
        column_class=Timestamp,
        params={
            "default": None,
            "null": True,
            "primary": False,
            "key": False,
            "unique": False,
            "index": False,
            "index_method": IndexMethod.btree,
        },
    )

    return manager



from __future__ import annotations

import itertools
import json
import typing as t
from collections import defaultdict
from functools import partial

import pydantic

from piccolo.columns import Column
from piccolo.columns.column_types import (
    JSON,
    JSONB,
    Array,
    Decimal,
    Email,
    ForeignKey,
    Numeric,
    Text,
    Timestamptz,
    Varchar,
)
from piccolo.table import Table
from piccolo.utils.encoding import load_json

try:
    from pydantic.config import JsonDict
except ImportError:
    JsonDict = dict  # type: ignore


def pydantic_json_validator(value: t.Optional[str], required: bool = True):
    if value is None:
        if required:
            raise ValueError("The JSON value wasn't provided.")
        else:
            return value

    try:
        load_json(value)
    except json.JSONDecodeError as e:
        raise ValueError("Unable to parse the JSON.") from e
    else:
        return value


def is_table_column(column: Column, table: t.Type[Table]) -> bool:
    """
    Verify that the given ``Column`` belongs to the given ``Table``.
    """
    if column._meta.table is table:
        return True
    elif (
        column._meta.call_chain
        and column._meta.call_chain[0]._meta.table is table
    ):
        # We also allow the column if it's joined from the table.
        return True
    return False


def validate_columns(
    columns: t.Tuple[Column, ...], table: t.Type[Table]
) -> bool:
    """
    Verify that each column is a ``Column``` instance, and its parent is the
    given ``Table``.
    """
    return all(
        isinstance(column, Column)
        and is_table_column(column=column, table=table)
        for column in columns
    )


def get_array_value_type(
    column: Array, inner: t.Optional[t.Type] = None
) -> t.Type:
    """
    Gets the correct type for an ``Array`` column (which might be
    multidimensional).
    """
    if isinstance(column.base_column, Array):
        inner_type = get_array_value_type(column.base_column, inner=inner)
    else:
        inner_type = get_pydantic_value_type(column.base_column)

    return t.List[inner_type]  # type: ignore


def get_pydantic_value_type(column: Column) -> t.Type:
    """
    Map the Piccolo ``Column`` to a Pydantic type.
    """
    value_type: t.Type

    if isinstance(column, (Decimal, Numeric)):
        value_type = pydantic.condecimal(
            max_digits=column.precision, decimal_places=column.scale
        )
    elif isinstance(column, Email):
        value_type = pydantic.EmailStr  # type: ignore
    elif isinstance(column, Varchar):
        value_type = pydantic.constr(max_length=column.length)
    elif isinstance(column, Array):
        value_type = get_array_value_type(column=column)
    else:
        value_type = column.value_type

    return value_type


def create_pydantic_model(
    table: t.Type[Table],
    nested: t.Union[bool, t.Tuple[ForeignKey, ...]] = False,
    exclude_columns: t.Tuple[Column, ...] = (),
    include_columns: t.Tuple[Column, ...] = (),
    include_default_columns: bool = False,
    include_readable: bool = False,
    all_optional: bool = False,
    model_name: t.Optional[str] = None,
    deserialize_json: bool = False,
    recursion_depth: int = 0,
    max_recursion_depth: int = 5,
    pydantic_config: t.Optional[pydantic.config.ConfigDict] = None,
    json_schema_extra: t.Optional[t.Dict[str, t.Any]] = None,
) -> t.Type[pydantic.BaseModel]:
    """
    Create a Pydantic model representing a table.

    :param table:
        The Piccolo ``Table`` you want to create a Pydantic serialiser model
        for.
    :param nested:
        Whether ``ForeignKey`` columns are converted to nested Pydantic models.
        If ``False``, none are converted. If ``True``, they all are converted.
        If a tuple of ``ForeignKey`` columns is passed in, then only those are
        converted.
    :param exclude_columns:
        A tuple of ``Column`` instances that should be excluded from the
        Pydantic model. Only specify ``include_columns`` or
        ``exclude_columns``.
    :param include_columns:
        A tuple of ``Column`` instances that should be included in the
        Pydantic model. Only specify ``include_columns`` or
        ``exclude_columns``.
    :param include_default_columns:
        Whether to include columns like ``id`` in the serialiser. You will
        typically include these columns in GET requests, but don't require
        them in POST requests.
    :param include_readable:
        Whether to include 'readable' columns, which give a string
        representation of a foreign key.
    :param all_optional:
        If True, all fields are optional. Useful for filters etc.
    :param model_name:
        By default, the classname of the Piccolo ``Table`` will be used, but
        you can override it if you want multiple Pydantic models based off the
        same Piccolo table.
    :param deserialize_json:
        By default, the values of any Piccolo ``JSON`` or ``JSONB`` columns are
        returned as strings. By setting this parameter to ``True``, they will
        be returned as objects.
    :param recursion_depth:
        Not to be set by the user - used internally to track recursion.
    :param max_recursion_depth:
        If using nested models, this specifies the max amount of recursion.
    :param pydantic_config:
        Allows you to configure some of Pydantic's behaviour. See the
        `Pydantic docs <https://docs.pydantic.dev/latest/api/config/#pydantic.config.ConfigDict>`_
        for more info.
    :param json_schema_extra:
        This can be used to add additional fields to the schema. This is
        very useful when using Pydantic's JSON Schema features. For example:

        .. code-block:: python

            >>> my_model = create_pydantic_model(Band, my_extra_field="Hello")
            >>> my_model.model_json_schema()
            {..., "my_extra_field": "Hello"}

    :returns:
        A Pydantic model.

    """  # noqa: E501
    if exclude_columns and include_columns:
        raise ValueError(
            "`include_columns` and `exclude_columns` can't be used at the "
            "same time."
        )

    if recursion_depth == 0:
        if exclude_columns:
            if not validate_columns(columns=exclude_columns, table=table):
                raise ValueError(
                    f"`exclude_columns` are invalid: {exclude_columns!r}"
                )

        if include_columns:
            if not validate_columns(columns=include_columns, table=table):
                raise ValueError(
                    f"`include_columns` are invalid: {include_columns!r}"
                )

    ###########################################################################

    columns: t.Dict[str, t.Any] = {}
    validators: t.Dict[str, t.Callable] = {}

    piccolo_columns = tuple(
        table._meta.columns
        if include_default_columns
        else table._meta.non_default_columns
    )

    if include_columns:
        include_columns_plus_ancestors = list(
            itertools.chain(
                include_columns, *[i._meta.call_chain for i in include_columns]
            )
        )
        piccolo_columns = tuple(
            i
            for i in piccolo_columns
            if any(
                i._equals(include_column)
                for include_column in include_columns_plus_ancestors
            )
        )

    if exclude_columns:
        piccolo_columns = tuple(
            i
            for i in piccolo_columns
            if not any(
                i._equals(exclude_column) for exclude_column in exclude_columns
            )
        )

    model_name = model_name or table.__name__

    for column in piccolo_columns:
        column_name = column._meta.name

        is_optional = True if all_optional else not column._meta.required

        #######################################################################
        # Work out the column type

        if isinstance(column, (JSON, JSONB)):
            if deserialize_json:
                value_type = pydantic.Json
            else:
                value_type = column.value_type
                validator = partial(
                    pydantic_json_validator, required=not is_optional
                )
                validators[
                    f"{column_name}_is_json"
                ] = pydantic.field_validator(column_name)(
                    validator  # type: ignore
                )
        else:
            value_type = get_pydantic_value_type(column=column)

        _type = t.Optional[value_type] if is_optional else value_type

        #######################################################################

        params: t.Dict[str, t.Any] = {}
        if is_optional:
            params["default"] = None

        if column._meta.db_column_name != column._meta.name:
            params["alias"] = column._meta.db_column_name

        extra: JsonDict = {
            "help_text": column._meta.help_text,
            "choices": column._meta.get_choices_dict(),
            "secret": column._meta.secret,
            "nullable": column._meta.null,
            "unique": column._meta.unique,
        }

        if isinstance(column, ForeignKey):
            if recursion_depth < max_recursion_depth and (
                (nested is True)
                or (
                    isinstance(nested, tuple)
                    and any(
                        column._equals(i)
                        for i in itertools.chain(
                            nested, *[i._meta.call_chain for i in nested]
                        )
                    )
                )
            ):
                nested_model_name = f"{model_name}.{column._meta.name}"
                _type = create_pydantic_model(
                    table=column._foreign_key_meta.resolved_references,
                    nested=nested,
                    include_columns=include_columns,
                    exclude_columns=exclude_columns,
                    include_default_columns=include_default_columns,
                    include_readable=include_readable,
                    all_optional=all_optional,
                    deserialize_json=deserialize_json,
                    recursion_depth=recursion_depth + 1,
                    max_recursion_depth=max_recursion_depth,
                    model_name=nested_model_name,
                )

            tablename = (
                column._foreign_key_meta.resolved_references._meta.tablename
            )
            target_column = (
                column._foreign_key_meta.resolved_target_column._meta.name
            )
            extra["foreign_key"] = {
                "to": tablename,
                "target_column": target_column,
            }

            if include_readable:
                columns[f"{column_name}_readable"] = (str, None)
        else:
            # This is used to tell Piccolo Admin that we want to display these
            # values using a specific widget.
            if isinstance(column, Text):
                extra["widget"] = "text-area"
            elif isinstance(column, (JSON, JSONB)):
                extra["widget"] = "json"
            elif isinstance(column, Timestamptz):
                extra["widget"] = "timestamptz"

            # It is useful for Piccolo API and Piccolo Admin to easily know
            # how many dimensions the array has.
            if isinstance(column, Array):
                extra["dimensions"] = column._get_dimensions()

        field = pydantic.Field(
            json_schema_extra={"extra": extra},
            **params,
        )

        columns[column_name] = (_type, field)

    pydantic_config = (
        pydantic_config.copy()
        if pydantic_config
        else pydantic.config.ConfigDict()
    )
    pydantic_config["arbitrary_types_allowed"] = True

    json_schema_extra_ = defaultdict(dict, **(json_schema_extra or {}))
    json_schema_extra_["extra"]["help_text"] = table._meta.help_text

    pydantic_config["json_schema_extra"] = dict(json_schema_extra_)

    model = pydantic.create_model(
        model_name,
        __config__=pydantic_config,
        __validators__=validators,
        **columns,
    )
    model.__qualname__ = model_name

    return model



from unittest import TestCase

from piccolo.schema import SchemaManager
from piccolo.table import Table
from tests.base import engines_skip


class Band(Table, schema="schema_1"):
    pass


@engines_skip("sqlite")
class TestListTables(TestCase):
    def setUp(self):
        Band.create_table().run_sync()

    def tearDown(self):
        Band.alter().drop_table().run_sync()

    def test_list_tables(self):
        """
        Make sure we can list all the tables in a schema.
        """
        schema_name = Band._meta.schema

        assert schema_name is not None
        table_list = (
            SchemaManager().list_tables(schema_name=schema_name).run_sync()
        )
        self.assertListEqual(table_list, [Band._meta.tablename])


@engines_skip("sqlite")
class TestCreateAndDrop(TestCase):
    def test_create_and_drop(self):
        """
        Make sure a schema can be created, and dropped.
        """
        manager = SchemaManager()

        # Make sure schema names with spaces, and clashing with keywords work.
        for schema_name in ("test_schema", "test schema", "user"):
            manager.create_schema(schema_name=schema_name).run_sync()

            self.assertIn(schema_name, manager.list_schemas().run_sync())

            manager.drop_schema(schema_name=schema_name).run_sync()
            self.assertNotIn(schema_name, manager.list_schemas().run_sync())


@engines_skip("sqlite")
class TestMoveTable(TestCase):
    new_schema = "schema_2"

    def setUp(self):
        Band.create_table(if_not_exists=True).run_sync()
        SchemaManager().create_schema(
            self.new_schema, if_not_exists=True
        ).run_sync()

    def tearDown(self):
        Band.alter().drop_table(if_exists=True).run_sync()
        SchemaManager().drop_schema(
            self.new_schema, if_exists=True, cascade=True
        ).run_sync()

    def test_move_table(self):
        """
        Make sure we can move a table to a different schema.
        """
        manager = SchemaManager()

        manager.move_table(
            table_name=Band._meta.tablename,
            new_schema=self.new_schema,
            current_schema=Band._meta.schema,
        ).run_sync()

        self.assertIn(
            Band._meta.tablename,
            manager.list_tables(schema_name=self.new_schema).run_sync(),
        )

        self.assertNotIn(
            Band._meta.tablename,
            manager.list_tables(schema_name="schema_1").run_sync(),
        )


@engines_skip("sqlite")
class TestRenameSchema(TestCase):
    manager = SchemaManager()
    schema_name = "test_schema"
    new_schema_name = "test_schema_2"

    def tearDown(self):
        for schema_name in (self.schema_name, self.new_schema_name):
            self.manager.drop_schema(
                schema_name=schema_name, if_exists=True
            ).run_sync()

    def test_rename_schema(self):
        """
        Make sure we can rename a schema.
        """
        self.manager.create_schema(
            schema_name=self.schema_name, if_not_exists=True
        ).run_sync()

        self.manager.rename_schema(
            schema_name=self.schema_name, new_schema_name=self.new_schema_name
        ).run_sync()

        self.assertIn(
            self.new_schema_name, self.manager.list_schemas().run_sync()
        )


@engines_skip("sqlite")
class TestDDL(TestCase):
    manager = SchemaManager()

    def test_create_schema(self):
        self.assertEqual(
            self.manager.create_schema(
                schema_name="schema_1", if_not_exists=False
            ).ddl,
            'CREATE SCHEMA "schema_1"',
        )

        self.assertEqual(
            self.manager.create_schema(
                schema_name="schema_1", if_not_exists=True
            ).ddl,
            'CREATE SCHEMA IF NOT EXISTS "schema_1"',
        )

    def test_drop_schema(self):
        self.assertEqual(
            self.manager.drop_schema(
                schema_name="schema_1", if_exists=False
            ).ddl,
            'DROP SCHEMA "schema_1"',
        )

        self.assertEqual(
            self.manager.drop_schema(
                schema_name="schema_1", if_exists=True
            ).ddl,
            'DROP SCHEMA IF EXISTS "schema_1"',
        )

        self.assertEqual(
            self.manager.drop_schema(
                schema_name="schema_1", if_exists=True, cascade=True
            ).ddl,
            'DROP SCHEMA IF EXISTS "schema_1" CASCADE',
        )

        self.assertEqual(
            self.manager.drop_schema(
                schema_name="schema_1", if_exists=False, cascade=True
            ).ddl,
            'DROP SCHEMA "schema_1" CASCADE',
        )

    def test_move_table(self):
        self.assertEqual(
            self.manager.move_table(
                table_name="band",
                new_schema="schema_2",
                current_schema="schema_1",
            ).ddl,
            'ALTER TABLE "schema_1"."band" SET SCHEMA "schema_2"',
        )

        self.assertEqual(
            self.manager.move_table(
                table_name="band",
                new_schema="schema_2",
            ).ddl,
            'ALTER TABLE "band" SET SCHEMA "schema_2"',
        )



import typing as t

from piccolo.columns.base import Column
from piccolo.columns.column_types import (
    Date,
    Integer,
    Time,
    Timestamp,
    Timestamptz,
)
from piccolo.querystring import QueryString

from .type_conversion import Cast

###############################################################################
# Postgres / Cockroach

ExtractComponent = t.Literal[
    "century",
    "day",
    "decade",
    "dow",
    "doy",
    "epoch",
    "hour",
    "isodow",
    "isoyear",
    "julian",
    "microseconds",
    "millennium",
    "milliseconds",
    "minute",
    "month",
    "quarter",
    "second",
    "timezone",
    "timezone_hour",
    "timezone_minute",
    "week",
    "year",
]


class Extract(QueryString):
    def __init__(
        self,
        identifier: t.Union[Date, Time, Timestamp, Timestamptz, QueryString],
        datetime_component: ExtractComponent,
        alias: t.Optional[str] = None,
    ):
        """
        .. note:: This is for Postgres / Cockroach only.

        Extract a date or time component from a ``Date`` / ``Time`` /
        ``Timestamp`` / ``Timestamptz`` column. For example, getting the month
        from a timestamp:

        .. code-block:: python

            >>> from piccolo.query.functions import Extract
            >>> await Concert.select(
            ...     Extract(Concert.starts, "month", alias="start_month")
            ... )
            [{"start_month": 12}]

        :param identifier:
            Identifies the column.
        :param datetime_component:
            The date or time component to extract from the column.

        """
        if datetime_component.lower() not in t.get_args(ExtractComponent):
            raise ValueError("The date time component isn't recognised.")

        super().__init__(
            f"EXTRACT({datetime_component} FROM {{}})",
            identifier,
            alias=alias,
        )


###############################################################################
# SQLite


class Strftime(QueryString):
    def __init__(
        self,
        identifier: t.Union[Date, Time, Timestamp, Timestamptz, QueryString],
        datetime_format: str,
        alias: t.Optional[str] = None,
    ):
        """
        .. note:: This is for SQLite only.

        Format a datetime value. For example:

        .. code-block:: python

            >>> from piccolo.query.functions import Strftime
            >>> await Concert.select(
            ...     Strftime(Concert.starts, "%Y", alias="start_year")
            ... )
            [{"start_month": "2024"}]

        :param identifier:
            Identifies the column.
        :param datetime_format:
            A string describing the output format (see SQLite's
            `documentation <https://www.sqlite.org/lang_datefunc.html>`_
            for more info).

        """
        super().__init__(
            f"strftime('{datetime_format}', {{}})",
            identifier,
            alias=alias,
        )


###############################################################################
# Database agnostic


def _get_engine_type(identifier: t.Union[Column, QueryString]) -> str:
    if isinstance(identifier, Column):
        return identifier._meta.engine_type
    elif isinstance(identifier, QueryString) and (
        columns := identifier.columns
    ):
        return columns[0]._meta.engine_type
    else:
        raise ValueError("Unable to determine the engine type")


def _extract_component(
    identifier: t.Union[Date, Time, Timestamp, Timestamptz, QueryString],
    sqlite_format: str,
    postgres_format: ExtractComponent,
    alias: t.Optional[str],
) -> QueryString:
    engine_type = _get_engine_type(identifier=identifier)

    return Cast(
        (
            Strftime(
                identifier=identifier,
                datetime_format=sqlite_format,
            )
            if engine_type == "sqlite"
            else Extract(
                identifier=identifier,
                datetime_component=postgres_format,
            )
        ),
        Integer(),
        alias=alias,
    )


def Year(
    identifier: t.Union[Date, Timestamp, Timestamptz, QueryString],
    alias: t.Optional[str] = None,
) -> QueryString:
    """
    Extract the year as an integer.
    """
    return _extract_component(
        identifier=identifier,
        sqlite_format="%Y",
        postgres_format="year",
        alias=alias,
    )


def Month(
    identifier: t.Union[Date, Timestamp, Timestamptz, QueryString],
    alias: t.Optional[str] = None,
) -> QueryString:
    """
    Extract the month as an integer.
    """
    return _extract_component(
        identifier=identifier,
        sqlite_format="%m",
        postgres_format="month",
        alias=alias,
    )


def Day(
    identifier: t.Union[Date, Timestamp, Timestamptz, QueryString],
    alias: t.Optional[str] = None,
) -> QueryString:
    """
    Extract the day as an integer.
    """
    return _extract_component(
        identifier=identifier,
        sqlite_format="%d",
        postgres_format="day",
        alias=alias,
    )


def Hour(
    identifier: t.Union[Time, Timestamp, Timestamptz, QueryString],
    alias: t.Optional[str] = None,
) -> QueryString:
    """
    Extract the hour as an integer.
    """
    return _extract_component(
        identifier=identifier,
        sqlite_format="%H",
        postgres_format="hour",
        alias=alias,
    )


def Minute(
    identifier: t.Union[Time, Timestamp, Timestamptz, QueryString],
    alias: t.Optional[str] = None,
) -> QueryString:
    """
    Extract the minute as an integer.
    """
    return _extract_component(
        identifier=identifier,
        sqlite_format="%M",
        postgres_format="minute",
        alias=alias,
    )


def Second(
    identifier: t.Union[Time, Timestamp, Timestamptz, QueryString],
    alias: t.Optional[str] = None,
) -> QueryString:
    """
    Extract the second as an integer.
    """
    return _extract_component(
        identifier=identifier,
        sqlite_format="%S",
        postgres_format="second",
        alias=alias,
    )


__all__ = (
    "Extract",
    "Strftime",
    "Year",
    "Month",
    "Day",
    "Hour",
    "Minute",
    "Second",
)



import typing as t

from piccolo.columns.column_types import Boolean
from piccolo.table import Table
from tests.base import TableTest


class MyTable(Table):
    boolean = Boolean(boolean=False, null=True)


class TestBoolean(TableTest):
    tables = [MyTable]

    def test_return_type(self) -> None:
        for value in (True, False, None, ...):
            kwargs: t.Dict[str, t.Any] = (
                {} if value is ... else {"boolean": value}
            )
            expected = MyTable.boolean.default if value is ... else value

            row = MyTable(**kwargs)
            row.save().run_sync()
            self.assertEqual(row.boolean, expected)

            row_from_db = (
                MyTable.select(MyTable.boolean)
                .where(
                    MyTable._meta.primary_key
                    == getattr(row, MyTable._meta.primary_key._meta.name)
                )
                .first()
                .run_sync()
            )
            assert row_from_db is not None

            self.assertEqual(
                row_from_db["boolean"],
                expected,
            )

    def test_eq_and_ne(self):
        """
        Make sure the `eq` and `ne` methods works correctly.
        """
        MyTable.insert(
            MyTable(boolean=True),
            MyTable(boolean=False),
            MyTable(boolean=True),
        ).run_sync()

        self.assertEqual(
            MyTable.count().where(MyTable.boolean.eq(True)).run_sync(), 2
        )

        self.assertEqual(
            MyTable.count().where(MyTable.boolean.ne(True)).run_sync(), 1
        )
